#  --*-- encoding: utf-8 --*--

# Nautilus-Meld
# Nautilus extension to invoke Meld for comparing selected files/folders
# ng.hong.quan@gmail.com
# Require Python >= 2.7

_VERSION = '0.1'

import os
import subprocess
import urllib
from collections import deque
from gi.repository import Nautilus, GObject

class DequeSet(deque):
	''' Data type like collections.deque, but only allow 1 occurence of each element '''
	def append(self, x):
		if len(self) and self[-1] == x:
			return
		if x in self:
			self.remove(x)
		super(DequeSet, self).append(x)

	def appendleft(self, x):
		if len(self) and self[0] == x:
			return
		if x in self:
			self.remove(x)
		super(DequeSet, self).appendleft(x)

	def extend(self, iterable):
		dup = [x for x in self if x in iterable]
		[self.remove(x) for x in dup]
		super(DequeSet, self).extend(iterable)

	def extendleft(self, iterable):
		dup = [x for x in self if x in iterable]
		[self.remove(x) for x in dup]
		super(DequeSet, self).extendlef(iterable)


class MeldExtension(GObject.GObject, Nautilus.MenuProvider):
	def __init__(self):
		try:
			meldbin = subprocess.check_output(['which', 'meld'])
			self.meldbin = meldbin.strip()
		except subprocess.CalledProcessError:
			self.meldbin = None
			return

		self.queue = DequeSet([], 3)

		# Create menu items, not attach to Nautilus menu yet
		self.item_meld = Nautilus.MenuItem(
			name='MeldExtension::meld_item',
			label='Compare in Meld',
			tip='Compare in Meld',
			icon='meld')
		self._makeitem_compare_each_other()
		self._makeitem_compare_with_last()
		self._makeitem_queue()
		self._makeitem_empty()

	def _makeitem_compare_each_other(self):
		self.item_comp_each = Nautilus.MenuItem(
			name='MeldExtension::compare_each',
			label='Compare each other',
			tip='Compare each other',
			icon='edit-copy')

	def _makeitem_compare_with_last(self):
		self.item_comp_last = Nautilus.MenuItem(
			name='MeldExtension::compare_last',
			label='Compare',
			tip='Compare',
			icon='edit-copy')

	def _makeitem_queue(self):
		self.item_queue = Nautilus.MenuItem(
			name='MeldExtension::add_file',
			label='Queue',
			tip='Add to queue',
			icon='document-import')

	def _makeitem_empty(self):
		self.item_empt = Nautilus.MenuItem(
			name='MeldExtension::empty_queue',
			label='Empty queue',
			tip='Empty queue',
			icon='edit-clear')

	def get_file_items(self, window, files):
		if self.meldbin is None:
			return

		qlen = len(self.queue)
		flen = len(files)
		if flen == 0 or flen + qlen > 3:
			return

		if not arelocal(files):
			return

		fls = [get_location(f) for f in files]
		meldmenu = self.attach_meldmenu()
		if qlen > 0:
			self.opt_compare_with_last(meldmenu, fls)
		elif flen > 1:
			self.opt_compare_each_other(meldmenu, fls)

		if qlen + flen < 3:
			self.opt_queue(meldmenu, fls)

		if qlen > 0:
			self.opt_empty(meldmenu)

		return (self.item_meld,)  # 1-tuple


	def opt_compare_each_other(self, menu, files):
		self.remove_old_handler(self.item_comp_each)
		hl = self.item_comp_each.connect('activate', self.launch_meld_batch, files)
		self.item_comp_each.handler = hl
		menu.append_item(self.item_comp_each)

	def opt_compare_with_last(self, menu, files):
		self.remove_old_handler(self.item_comp_last)
		last = self.queue[-1]
		last = os.path.basename(last)
		label = 'Compare with {0}'.format(mid_ellipsis(last).replace('_', '__'))
		self.item_comp_last.set_property('label', label)
		self.item_comp_last.set_property('tip', 'Compare with {0}'.format(last))
		hl = self.item_comp_last.connect('activate', self.add_to_meld, files)
		self.item_comp_last.handler = hl
		menu.append_item(self.item_comp_last)

	def opt_queue(self, menu, files):
		self.remove_old_handler(self.item_queue)
		if len(files) == 1:
			f = os.path.basename(files[0])
			self.item_queue.set_property('label', 'Queue {0}'.format(mid_ellipsis(f).replace('_', '__')))
			self.item_queue.set_property('tip', 'Add {0} to queue'.format(f))
		else:
			self.item_queue.set_property('label', 'Queue these files')
			self.item_queue.set_property('tip', 'Add these files to queue')

		hl = self.item_queue.connect('activate', self.add_files, files)
		self.item_queue.handler = hl
		menu.append_item(self.item_queue)

	def opt_empty(self, menu):
		self.remove_old_handler(self.item_empt)
		hl = self.item_empt.connect('activate', self.empty_queue)
		self.item_empt.handler = hl
		menu.append_item(self.item_empt)

	def add_file(self, menu, file):
		self.queue.append(file)

	def add_files(self, menu, files):
		self.queue.extend(files)

	def empty_queue(self, menu):
		self.queue.clear()

	def add_to_meld(self, menu, files):
		self.queue.extend(files)
		opts = ' '.join([wrap_filename(f) for f in self.queue])
		self.queue.clear()
		comm = '{0} {1} &'.format(self.meldbin, opts)
		os.system(comm)

	def launch_meld_batch(self, menu, files):
		opt = ' '.join([wrap_filename(f) for f in files])
		self.queue.clear()
		comm = '{0} {1} &'.format(self.meldbin, opt)
		print(comm)
		os.system(comm)

	def attach_meldmenu(self):
		menu = Nautilus.Menu()
		self.item_meld.set_submenu(menu)
		return menu

	def remove_old_handler(self, item):
		try:
			ohl = item.handler
		except AttributeError:
			return
		if ohl:
			item.disconnect(ohl)


#----  Functions  ----#

MAXCHAR = 16

def arelocal(files):
	for f in files:
		if not f.get_uri().startswith("file://"):
			return False
	return True

def escape_quote(string):
	return string.replace('"', '\\"')

def get_location(file):
	return urllib.unquote(file.get_uri()[7:])

def mid_ellipsis(strg, length=MAXCHAR):
	if len(strg) > length:
		mid = length/2
		return strg[:mid] + '…' + strg[-(length-mid-1):]
	else:
		return strg

def wrap_filename(fn):
	return '"{0}"'.format(escape_quote(fn))